use super::{
    VerbContext, UserVerb, UserVerbRef, UResult, UserError, user_error,
    get_player_item_or_fail,
    look,
    open::{DoorSituation, is_door_in_direction, attempt_open_immediate},
};
use async_trait::async_trait;
use crate::{
    DResult,
    language,
    regular_tasks::queued_command::{
        QueueCommandHandler,
        QueueCommand,
        queue_command
    },
    static_content::{
        room::{self, Direction, ExitType, ExitClimb, MaterialType},
        dynzone::{dynzone_by_type, ExitTarget as DynExitTarget, DynzoneType},
    },
    models::{
        item::{
            Item,
            ItemSpecialData,
            SkillType,
            LocationActionType,
            DoorState,
            ActiveClimb,
        },
        consent::ConsentType,
    },
    services::{
        comms::broadcast_to_room,
        skills::skill_check_and_grind,
        combat::{
            stop_attacking_mut,
            handle_resurrect,
            change_health
        },
        check_consent,
    }
};
use std::sync::Arc;
use mockall_double::double;
#[double] use crate::db::DBTrans;
use std::time;
use ansi::ansi;
use rand_distr::{Normal, Distribution};

pub async fn announce_move(trans: &DBTrans, character: &Item, leaving: &Item, arriving: &Item) -> DResult<()> {
    let msg_leaving_exp = format!("{} departs towards {}\n",
                                  &character.display_for_sentence(true, 1, true),
                                  &arriving.display);
    let msg_leaving_nonexp = format!("{} departs towards {}\n",
                                     character.display_for_sentence(true, 1, false),
                                     arriving.display_less_explicit
                                       .as_ref()
                                       .unwrap_or(&arriving.display));
    broadcast_to_room(trans, &format!("{}/{}", &leaving.item_type, &leaving.item_code),
                      None, &msg_leaving_exp, Some(&msg_leaving_nonexp)).await?;
    
    let msg_arriving_exp = format!("{} arrives from {}\n", &character.display_for_sentence(true, 1, true),
                                   &leaving.display);
    let msg_arriving_nonexp = format!("{} arrives from {}\n",
                                     character.display_for_sentence(true, 1, false),
                                     leaving.display_less_explicit
                                       .as_ref()
                                       .unwrap_or(&leaving.display));
    broadcast_to_room(trans, &format!("{}/{}", &arriving.item_type, &arriving.item_code),
                      None, &msg_arriving_exp, Some(&msg_arriving_nonexp)).await?;
    Ok(())
}

async fn move_to_where(
    trans: &DBTrans,
    use_location: &str,
    direction: &Direction,
    mover_for_exit_check: Option<&mut Item>,
    player_ctx: &mut Option<&mut VerbContext<'_>>
) -> UResult<(String, Option<Item>, Option<&'static ExitClimb>)> {
    // Firstly check dynamic exits, since they apply to rooms and dynrooms...
    if let Some(dynroom_result) = trans.find_exact_dyn_exit(use_location, direction).await? {
        return Ok((format!("{}/{}",
                           &dynroom_result.item_type,
                           &dynroom_result.item_code), Some(dynroom_result), None));
    }
    
    let (heretype, herecode) = use_location.split_once("/").unwrap_or(("room", "repro_xv_chargen"));

    if heretype == "dynroom" {
        let old_dynroom_item = match trans.find_item_by_type_code(heretype, herecode).await? {
            None => user_error("Your current room has vanished!".to_owned())?,
            Some(v) => v
        };
        let (dynzone_code, dynroom_code) = match old_dynroom_item.special_data.as_ref() {
            Some(ItemSpecialData::DynroomData { dynzone_code, dynroom_code }) => (dynzone_code, dynroom_code),
            _ => user_error("Your current room is invalid!".to_owned())?
        };
        let dynzone = dynzone_by_type().get(&DynzoneType::from_str(dynzone_code)
                                            .ok_or_else(|| UserError("The type of your current zone no longer exists".to_owned()))?)
              .ok_or_else(|| UserError("The type of your current zone no longer exists".to_owned()))?;
        let dynroom = dynzone.dyn_rooms.get(dynroom_code.as_str())
            .ok_or_else(|| UserError("Your current room type no longer exists".to_owned()))?;
        let exit = dynroom.exits.iter().find(|ex| ex.direction == *direction)
            .ok_or_else(|| UserError("There is nothing in that direction".to_owned()))?;
        return match exit.target {
            DynExitTarget::ExitZone => {
                let (zonetype, zonecode) = old_dynroom_item.location.split_once("/")
                    .ok_or_else(|| UserError("Invalid zone for your room".to_owned()))?;
                let zoneitem = trans.find_item_by_type_code(zonetype, zonecode).await?
                    .ok_or_else(|| UserError("Can't find your zone".to_owned()))?;
                let zone_exit = match zoneitem.special_data.as_ref() {
                    Some(ItemSpecialData::DynzoneData { zone_exit: None, .. }) =>
                        user_error("That exit doesn't seem to go anywhere".to_owned())?,
                    Some(ItemSpecialData::DynzoneData { zone_exit: Some(zone_exit), .. }) => zone_exit,
                    _ => user_error("The zone you are in has invalid data associated with it".to_owned())?,
                };
                Ok((zone_exit.to_string(), None, None))
            },
            DynExitTarget::Intrazone { subcode } => {
                let to_item = trans.find_item_by_location_dynroom_code(&old_dynroom_item.location, &subcode).await?
                    .ok_or_else(|| UserError("Can't find the room in that direction.".to_owned()))?;
                Ok((format!("{}/{}", &to_item.item_type, &to_item.item_code), Some(to_item), None))
            }
        }
    }
    if heretype != "room" {
        user_error("Navigating outside rooms not yet supported.".to_owned())?
    }
    let room = room::room_map_by_code().get(herecode)
        .ok_or_else(|| UserError("Can't find your current location".to_owned()))?;
    let exit = room.exits.iter().find(|ex| ex.direction == *direction)
        .ok_or_else(|| UserError("There is nothing in that direction".to_owned()))?;

    match exit.exit_type {
        ExitType::Free => {}
        ExitType::Blocked(blocker) => {
            if let Some(mover) = mover_for_exit_check {
                if let Some(ctx) = player_ctx {
                    if !blocker.attempt_exit(*ctx, mover, exit).await? {
                        user_error("Stopping movement".to_owned())?;
                    }
                }
            }
        }
    }

    let new_room =
        room::resolve_exit(room, exit).ok_or_else(|| UserError("Can't find that room".to_owned()))?;
    Ok((format!("room/{}", new_room.code), None, exit.exit_climb.as_ref()))
}

pub async fn check_room_access(trans: &DBTrans, player: &Item, room: &Item) -> UResult<()> {
    let (owner_t, owner_c) = match room.owner.as_ref().and_then(|o| o.split_once("/")) {
        None => return Ok(()),
        Some(v) => v
    };
    if owner_t == &player.item_type && owner_c == &player.item_code {
        return Ok(());
    }
    let owner = match trans.find_item_by_type_code(owner_t, owner_c).await? {
        None => return Ok(()),
        Some(v) => v
    };

    if check_consent(trans, "enter", &ConsentType::Visit, player, &owner).await? {
        return Ok(());
    }

    let mut player_hypothet = (*player).clone();
    // We are asking hypothetically if they entered the room, could they fight
    // the owner? We won't save this yet.
    player_hypothet.location = room.refstr();
    if check_consent(trans, "enter", &ConsentType::Fight, &player_hypothet, &owner).await? {
        return Ok(());
    }
    
    user_error(ansi!("<yellow>Your wristpad buzzes and your muscles lock up, stopping you entering.<reset> \
                      It seems this is private property and you haven't been invited here with \
                      <bold>allow visit<reset>, nor do you have a <bold>allow fight<reset> in force with \
                      the owner here.").to_owned())?
}

pub async fn handle_fall(
    trans: &DBTrans,
    faller: &mut Item,
    fall_dist: u64
) -> UResult<String> {
    // TODO depend on distance, armour, etc...
    // This is deliberately less damage than real life for the distance,
    // since we'll assume the wristpad provides reflexes to buffer some damage.

    let damage_modifier = match faller.location.split_once("/") {
        Some((ltype, lcode)) if ltype == "room" => {
            match room::room_map_by_code().get(lcode) {
                None => 1.0,
                Some(room) => match room.material_type {
                    MaterialType::WaterSurface | MaterialType::Underwater => {
                        return Ok("lands with a splash".to_owned());
                    },
                    MaterialType::Soft { damage_modifier } => damage_modifier,
                    MaterialType::Normal => 1.0
                }
            }
        },
        _ => 1.0
    };

    let modified_safe_distance = 5.0 / damage_modifier;
    if (fall_dist as f64) < modified_safe_distance {
        return Ok("lands softly".to_owned());
    }
    // The force is proportional to the square root of the fall distance.
    let damage = ((fall_dist as f64 - modified_safe_distance).sqrt() * 3.0 * damage_modifier *
        Normal::new(1.0, 0.3)?
        .sample(&mut rand::thread_rng())) as i64;

    if damage > 0 {
        change_health(trans, -damage, faller, "You fell", "You fell").await?;
    }
    
    let descriptor = if damage >= 30 {
        "smashes violently into the ground like a comet with a massive boom"
    } else if damage >= 25 {
        "smashes violently into the ground with a very loud bang"
    } else if damage >= 20 {
        "smashes violently into the ground with a loud bang"
    }  else if damage >= 15 {
        "smashes into the ground with a loud thump"
    } else if damage >= 10 {
        "smashes into the ground with a thump"
    } else {
        "lands with a thump"
    };
    Ok(descriptor.to_owned())
}

pub async fn attempt_move_immediate(
    trans: &DBTrans,
    orig_mover: &Item,
    direction: &Direction,
    // player_ctx should only be Some if called from queue_handler finish_command
    // for the orig_mover's queue, because might re-queue a move command.
    mut player_ctx: &mut Option<&mut VerbContext<'_>>
) -> UResult<()> {
    let use_location = if orig_mover.death_data.is_some() {
        if orig_mover.item_type != "player" {
            user_error("Dead players don't move".to_owned())?;
        }
        "room/repro_xv_respawn"
    } else {
        &orig_mover.location
    };

    match is_door_in_direction(trans, direction, use_location).await? {
        DoorSituation::NoDoor |
        DoorSituation::DoorOutOfRoom { state: DoorState { open: true, .. }, .. } => {},
        DoorSituation::DoorIntoRoom { state: DoorState { open: true, .. }, room_with_door, .. } => {
            check_room_access(trans, orig_mover, &room_with_door).await?;
        }
        _ => {
            attempt_open_immediate(trans, player_ctx, orig_mover, direction).await?;
            match player_ctx.as_mut() {
                None => {
                    // NPCs etc... open and move in one step, but can't unlock.
                },
                Some(actual_player_ctx) => {
                    // Players take an extra step. So tell them to come back.
                    actual_player_ctx.session_dat.queue.push_front(
                        QueueCommand::Movement { direction: direction.clone() }
                    );
                    return Ok(());
                }
            }
        }
    }
    
    let mut mover = (*orig_mover).clone();
    let (new_loc, new_loc_item, climb_opt) =
        move_to_where(trans, use_location, direction, Some(&mut mover), &mut player_ctx).await?;

    let mut skip_escape_check: bool = false;
    let mut escape_check_only: bool = false;
    
    if let Some(climb) = climb_opt {
        // We need to think about if NPCs climb at all.
        if let Some(ctx) = player_ctx {
            if let Some(active_climb) = mover.active_climb.clone() {
                skip_escape_check = true; // Already done if we get here.
                let skills = skill_check_and_grind(trans, &mut mover, &SkillType::Climb,
                                                   climb.difficulty as f64).await?;
                let mut narrative = String::new();
                if skills <= -0.25 {
                    // Crit fail - they have fallen.
                    let (fall_dist, from_room, to_room) = if climb.height < 0 {
                        // At least they get to where they want to go!
                        mover.location = new_loc.clone();
                        (climb.height.abs() as u64 - active_climb.height, new_loc.to_owned(), use_location.to_owned())
                    } else {
                        (active_climb.height, use_location.to_owned(), new_loc.to_owned())
                    };
                    mover.active_climb = None;
                    let descriptor = handle_fall(&trans, &mut mover, fall_dist).await?;
                    let msg_exp = format!(
                        "{} loses {} grip from {} metres up and {}!\n",
                        &mover.display_for_sentence(true, 1, true),
                        &mover.pronouns.possessive,
                        fall_dist,
                        &descriptor
                    );
                    let msg_nonexp = format!(
                        "{} loses {} grip from {} metres up and {}!\n",
                        &mover.display_for_sentence(true, 1, false),
                        &mover.pronouns.possessive,
                        fall_dist,
                        &descriptor
                    );
                    trans.save_item_model(&mover).await?;
                    broadcast_to_room(&trans, &from_room,
                                      None, &msg_exp, Some(&msg_nonexp)).await?;
                    broadcast_to_room(&trans, &to_room,
                                      None, &msg_exp, Some(&msg_nonexp)).await?;
                    ctx.session_dat.queue.truncate(0);
                    return Ok(());
                } else if skills <= 0.0 {
                    if climb.height >= 0 {
                        narrative.push_str("You lose your grip and slide a metre back down");
                    } else {
                        narrative.push_str("You struggle to find a foothold and reluctantly climb a metre back up");
                    }
                    if let Some(ac) = mover.active_climb.as_mut() {
                        if ac.height > 0 {
                            ac.height -= 1;
                        }
                    }
                } else {
                    if climb.height < 0 {
                        narrative.push_str("You climb down another metre");
                    } else {
                        narrative.push_str("You climb up another metre");
                    }
                    if let Some(ac) = mover.active_climb.as_mut() {
                        ac.height += 1;
                    }
                }
                if let Some(ac) = mover.active_climb.as_ref() {
                    if climb.height >= 0 && ac.height >= climb.height as u64 {
                        trans.queue_for_session(&ctx.session,
                                                Some("You brush yourself off and finish climbing - you \
                                                      made it to the top!\n")).await?;
                        mover.active_climb = None;
                    } else if climb.height < 0 && ac.height >= (-climb.height) as u64 {
                        trans.queue_for_session(&ctx.session,
                                                Some("You brush yourself off and finish climbing - you \
                                                      made it down!\n")).await?;
                        mover.active_climb = None;
                    } else {
                        let progress_quant = (((ac.height as f64) / (climb.height.abs() as f64)) * 10.0) as u64;
                        trans.queue_for_session(
                            &ctx.session,
                            Some(&format!(ansi!("<bold>[<reset><cyan>{}{}<reset><bold>] [<reset>{}/{} m<bold>]<reset> {}\n"),
                                          "=".repeat(progress_quant as usize), " ".repeat((10 - progress_quant) as usize),
                                          ac.height, climb.height.abs(), &narrative
                            ))).await?;
                        ctx.session_dat.queue.push_front(
                            QueueCommand::Movement { direction: direction.clone() }
                        );
                        trans.save_item_model(&mover).await?;        
                        return Ok(());
                    }
                }
            } else {
                let msg_exp = format!("{} starts climbing {}\n",
                                      &orig_mover.display_for_sentence(true, 1, true),
                                      &direction.describe_climb(if climb.height > 0 { "up" } else { "down" }));
                let msg_nonexp = format!("{} starts climbing {}\n",
                                         &orig_mover.display_for_sentence(true, 1, false),
                                         &direction.describe_climb(if climb.height > 0 { "up" } else { "down" }));
                broadcast_to_room(&trans, &use_location,
                                  None, &msg_exp, Some(&msg_nonexp)).await?;
                
                mover.active_climb = Some(ActiveClimb { ..Default::default() });
                
                ctx.session_dat.queue.push_front(
                    QueueCommand::Movement { direction: direction.clone() }
                );
                escape_check_only = true;
            }
        } else {
            user_error("NPC climbing not supported yet.".to_owned())?;
        }
    }

    if !skip_escape_check {
        match mover.active_combat.as_ref().and_then(|ac| ac.attacking.clone()) {
            None => {}
            Some(old_victim) => {
                if let Some((vcode, vtype)) = old_victim.split_once("/") {
                    if let Some(vitem) = trans.find_item_by_type_code(vcode, vtype).await? {
                        let mut vitem_mut = (*vitem).clone();
                        stop_attacking_mut(trans, &mut mover, &mut vitem_mut, false).await?;
                        trans.save_item_model(&vitem_mut).await?
                    }
                }
            }
        }
        match mover.active_combat.clone().as_ref().map(|ac| &ac.attacked_by[..]) {
            None | Some([]) => {}
            Some(attackers) => {
                let mut attacker_names = Vec::new();
                let mut attacker_items = Vec::new();
                if let Some(ctx) = player_ctx.as_ref() {
                    for attacker in &attackers[..] {
                        if let Some((acode, atype)) = attacker.split_once("/") {
                            if let Some(aitem) = trans.find_item_by_type_code(acode, atype).await? {
                                attacker_names.push(aitem.display_for_session(ctx.session_dat));
                                attacker_items.push(aitem);
                            }
                        }
                    }
                }
                let attacker_names_ref = attacker_names.iter().map(|n| n.as_str()).collect::<Vec<&str>>();
                let attacker_names_str = language::join_words(&attacker_names_ref[..]);
                if skill_check_and_grind(trans, &mut mover, &SkillType::Dodge, attackers.len() as f64 + 8.0).await? >= 0.0 {
                    if let Some(ctx) = player_ctx.as_ref() {
                        trans.queue_for_session(ctx.session,
                                                Some(&format!("You successfully get away from {}\n",
                                                              &attacker_names_str))).await?;
                    }
                    for item in &attacker_items[..] {
                        let mut item_mut = (**item).clone();
                        stop_attacking_mut(trans, &mut item_mut, &mut mover, true).await?;
                        trans.save_item_model(&item_mut).await?;
                    }
                } else {
                    if let Some(ctx) = player_ctx.as_ref() {
                        trans.queue_for_session(ctx.session,
                                                Some(&format!("You try and fail to run past {}\n",
                                                              &attacker_names_str))).await?;
                    }
                    trans.save_item_model(&mover).await?;
                    return Ok(());
                }
            }
        }
    }
    if escape_check_only {
        trans.save_item_model(&mover).await?;        
        return Ok(());
    }
    
    if mover.death_data.is_some() {
        if !handle_resurrect(trans, &mut mover).await? {
            user_error("You couldn't be resurrected.".to_string())?;
        }
    }
    
    mover.location = new_loc.clone();
    mover.action_type = LocationActionType::Normal;
    mover.active_combat = None;

    trans.save_item_model(&mover).await?;

    if let Some(ctx) = player_ctx {
        look::VERB.handle(ctx, "look", "").await?;
    }

    if let Some((old_loc_type, old_loc_code)) = use_location.split_once("/") {
        if let Some(old_room_item) = trans.find_item_by_type_code(old_loc_type, old_loc_code).await? {
            if let Some((new_loc_type, new_loc_code)) = new_loc.split_once("/") {
                if let Some(new_room_item) = match new_loc_item {
                    None => trans.find_item_by_type_code(new_loc_type, new_loc_code).await?,
                    v => v.map(Arc::new)
                } {
                    announce_move(&trans, &mover, &old_room_item, &new_room_item).await?;
                }
            }
        }
    }

    Ok(())
}

pub struct QueueHandler;
#[async_trait]
impl QueueCommandHandler for QueueHandler {
    async fn start_command(&self, _ctx: &mut VerbContext<'_>, _command: &QueueCommand)
                           -> UResult<time::Duration> {    
        Ok(time::Duration::from_secs(1))
    }

    #[allow(unreachable_patterns)]
    async fn finish_command(&self, ctx: &mut VerbContext<'_>, command: &QueueCommand)
                            -> UResult<()> {
        let direction = match command {
            QueueCommand::Movement { direction } => direction,
            _ => user_error("Unexpected command".to_owned())?
        };
        let player_item = get_player_item_or_fail(ctx).await?;
        attempt_move_immediate(ctx.trans, &player_item, direction, &mut Some(ctx)).await?;
        Ok(())
    }
}

pub struct Verb;

#[async_trait]
impl UserVerb for Verb {
    async fn handle(self: &Self, ctx: &mut VerbContext, verb: &str, remaining: &str) -> UResult<()> {
        let dir = Direction::parse(
            &(verb.to_owned() + " " + remaining.trim()).trim())
            .ok_or_else(|| UserError("Unknown direction".to_owned()))?;
        queue_command(ctx, &QueueCommand::Movement { direction: dir.clone() }).await
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
